#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

from abc import ABC, abstractmethod

import numpy as np
from scipy.special import gamma as gamma_func

from ciderpress.dft.feat_normalizer import (
    ConstantNormalizer,
    DensityNormalizer,
    FeatNormalizerList,
    get_normalizer_from_exponent_params,
)

LDA_FACTOR = -3.0 / 4.0 * (3.0 / np.pi) ** (1.0 / 3)
CFC = (3.0 / 10) * (3 * np.pi**2) ** (2.0 / 3)
ALPHA_TOL = 1e-10


def _check_l1_dots(l1_dots, nl1):
    for dot in l1_dots:
        if not isinstance(dot, (tuple, list)) or len(dot) != 2:
            raise ValueError("l1_dots items must be 2-tuples")
        for d in dot:
            if not (-1 <= d < nl1):
                raise ValueError("Illegal index in feat_dots")


def _get_fl_ueg(s):
    tol = 1e-10
    # first, handle the numerically unstable cases
    if s < 0:
        sign = -1
    else:
        sign = 1
    if np.abs(s + 0.5) < tol:
        return 1.5
    if np.abs(s) < tol:
        return 1.0
    elif np.abs(s - 0.5) < tol:
        return 0.75
    elif np.abs(s - 1.0) < tol:
        return 0.6
    # fmt: off
    return (
        sign * 4 * np.pi * 3 * np.cos(np.pi * s)
        * gamma_func(-1 - 2 * s) / (3 + 2 * s) * 4**s
        * gamma_func(1.5 + s)
        / (np.pi**1.5 * np.abs(gamma_func(-s)))
    )
    # fmt: on


def get_cider_exponent(
    rho, sigma, tau, a0=1.0, grad_mul=0.0, tau_mul=0.03125, rhocut=1e-10, nspin=1
):
    if nspin > 2:
        raise ValueError
    if isinstance(rho, np.ndarray):
        isarray = True
    else:
        isarray = False
        rho = np.asarray([rho], dtype=np.float64)
        sigma = np.asarray([sigma], dtype=np.float64)
        tau = np.asarray([tau], dtype=np.float64)
    tau_fac = tau_mul * 1.2 * (6 * np.pi**2) ** (2.0 / 3) / np.pi
    cond = rho < rhocut
    rho = rho.copy()
    rho[cond] = rhocut
    if nspin == 1:
        B = np.pi / 2 ** (2.0 / 3) * (a0 - tau_fac)
    else:
        B = np.pi * (a0 - tau_fac)
    C = np.pi / 2 ** (2.0 / 3) * tau_fac / CFC
    ascale = B * rho ** (2.0 / 3) + C * tau / rho
    dadrho = 2 * B / (3 * rho ** (1.0 / 3)) - (C * tau / rho) / rho
    dadtau = C / rho
    if grad_mul != 0:
        grad_fac = grad_mul * 1.2 * (6 * np.pi**2) ** (2.0 / 3) / np.pi
        grad_fac *= np.pi / 2 ** (2.0 / 3) * 0.125 / CFC
        ascale += grad_fac * sigma / (rho * rho)
        dadsigma = grad_fac / (rho * rho)
        dadrho -= 2 * grad_fac * sigma / (rho * rho * rho)
    else:
        dadsigma = np.zeros_like(ascale)
    dadrho[cond] = 0
    dadsigma[cond] = 0
    dadtau[cond] = 0
    if not isarray:
        ascale = ascale.item()
        dadrho = dadrho.item()
        dadsigma = dadsigma.item()
        dadtau = dadtau.item()
    return ascale, dadrho, dadsigma, dadtau


def get_cider_exponent_gga(rho, sigma, a0=1.0, grad_mul=0.03125, rhocut=1e-10, nspin=1):
    if nspin > 2:
        raise ValueError
    if isinstance(rho, np.ndarray):
        isarray = True
    else:
        isarray = False
        rho = np.asarray([rho], dtype=np.float64)
        sigma = np.asarray([sigma], dtype=np.float64)
    cond = rho < rhocut
    rho = rho.copy()
    rho[cond] = rhocut
    if nspin == 1:
        B = np.pi / 2 ** (2.0 / 3) * a0
    else:
        B = np.pi * a0
    ascale = B * rho ** (2.0 / 3)
    dadrho = 2 * B / (3 * rho ** (1.0 / 3))
    if grad_mul != 0:
        grad_fac = grad_mul * 1.2 * (6 * np.pi**2) ** (2.0 / 3) / np.pi
        grad_fac *= np.pi / 2 ** (2.0 / 3) * 0.125 / CFC
        ascale += grad_fac * sigma / (rho * rho)
        dadsigma = grad_fac / (rho * rho)
        dadrho -= 2 * grad_fac * sigma / (rho * rho * rho)
    else:
        dadsigma = np.zeros_like(ascale)
    dadrho[cond] = 0
    dadsigma[cond] = 0
    if not isarray:
        ascale = ascale.item()
        dadrho = dadrho.item()
        dadsigma = dadsigma.item()
    return ascale, dadrho, dadsigma


def _get_ueg_expnt(aval, tval, rho):
    return get_cider_exponent(
        rho,
        0,
        CFC * rho ** (5.0 / 3),
        a0=aval,
        grad_mul=0,
        tau_mul=tval,
        rhocut=0,
        nspin=1,
    )[0]


class BaseSettings(ABC):
    @property
    @abstractmethod
    def nfeat(self):
        """
        Returns:
            Number of features in this feature set.
        """

    @property
    def is_empty(self):
        return self.nfeat == 0

    @abstractmethod
    def get_feat_usps(self):
        """
        Returns:
            A list of floats with the uniform scaling power for each feature.
        """

    @abstractmethod
    def ueg_vector(self, rho=1.0):
        """
        Return the feature vector for the uniform electron gas
        as a function of the density rho

        Args:
            rho (float): Value of the density

        Returns:
            np.ndarray(float): UEG feature vector
        """

    def get_reasonable_normalizer(self):
        """
        Returns:
            norms (list[FeatNormalizer]): list of normalizers for these features
        """


class EmptySettings(BaseSettings):
    @property
    def nfeat(self):
        return 0

    def get_feat_usps(self):
        return []

    def ueg_vector(self, rho=1.0):
        return np.array([])

    def get_reasonable_normalizer(self):
        return []


class FracLaplSettings(BaseSettings):
    """
    Settings for Fractional Laplacian feature set.
    """

    def __init__(self, slist, nk0, nk1, l1_dots, nd1=0, ld_dots=None, ndd=0):
        """
        Initialize FracLaplSettings object. Note, in the documentation below
        on the indexing of the features, fl_rho is the part of the density
        ingredient vector AFTER the semilocal part.

        Note on the l=1 feature indexing. If l1_dots[i] = (j, k), then::

            fl_feat[i+nk0] = einsum(
                'xg,xg->g',
                fl_rho[3*j+nk0 : 3*j+3+nk0],
                fl_rho[3*k+nk0 : 3*k+3+nk0],
            )

        If ld_dots[i] = (j, k), then::

            nstart = nk0 + 3 * nk1
            fl_feat[i + nk0 + len(l1_dots)] = einsum(
                'xg,xg->g',
                fl_rho[3*j+nstart : 3*j+3+nstart],
                fl_rho[3*k+nstart : 3*k+3+nstart],
            )
            nstart = nk0 + 3 * nk1 + 3 * nd1
            fl_feat[i + nk0 + len(l1_dots) + len(ld_dots)] = fl_rho[i + nstart]

        Args:
            slist (list of float): s parameters for the fractional
                Laplacian. For each s, :math:`(-\\Delta)^s \\phi_i` is computed
                for each single-particle orbital.
            nk0 (int): Number of scalar (l=0) features. Must be <= len(slist).
                fl_feat[i] = fl_rho[i] = :math:`(-\\Delta')^s \\rho(r, r') |_{r'=r}`,
                with s=slist[i].
            nk1 (int): Number of vector (l=1) features F_s^1. Must be <= len(slist).
                fl_rho[3*i+nk0 : 3*i+3+nk0] =
                :math:`\\nabla'' (-\\Delta')^s \\rho(r'', r') |_{r'=r,r''=r}`,
                with s=slist[i].
            l1_dots (list of (int, int)): List of index tuples for contracting
                the l=1 F_s^1 features. -1 indicates to use the semilocal
                density gradient.
            nd1 (int): Number of vector features F_s^d. Must be ``<= len(slist)``
                With ``nstart = nk0 + 3 * nk1``,
                ``fl_rho[i * 3 + nstart : (i+1) * 3 + nstart] =``
                :math:`\\nabla' (-\\Delta')^s \\rho(r, r') |_{r'=r}`,
                with s=slist[i].
            ld_dots (list of (int, int)): Same as l1_dots but for the F_s^d
                features. -1 indicates to use the semilocal density gradient.
            ndd (int): Numer of dot product features :math:`F_s^{dd}`. Must be <= nd1
        """
        self.slist = slist
        assert nk0 <= self.npow
        assert nk1 <= self.npow
        assert nd1 <= self.npow
        assert ndd <= nd1
        self.l1_dots = l1_dots
        self._ndd = ndd
        self._nd1 = nd1
        if ld_dots is None:
            ld_dots = []
        self._ld_dots = ld_dots
        self._nk0 = nk0
        self._nk1 = nk1
        ir = 0
        self.a0_inds = []
        self.a1_inds = []
        for ik in range(self.nk0):
            self.a0_inds.append(ir)
            ir += 1
        for ik in range(self.nk1):
            self.a1_inds.append(ir)
            ir += 3
        ir += 3 * self.nd1 + self.ndd
        self._size = ir
        _check_l1_dots(l1_dots, self.nk1)

    @property
    def npow(self):
        return len(self.slist)

    @property
    def size(self):
        return self._size

    @property
    def nd1(self):
        return self._nd1

    @property
    def ndd(self):
        return self._ndd

    def get_ik0(self, ind):
        return ind + 3 * min(self.nd1, ind)

    @property
    def nk0(self):
        return self._nk0

    @property
    def nk1(self):
        return self._nk1

    @property
    def nrho(self):
        return self.nk0 + 3 * self.nk1 + 3 * self.nd1 + self.ndd

    @property
    def ld_dots(self):
        return self._ld_dots

    @property
    def nfeat(self):
        return self.nk0 + len(self.l1_dots) + len(self.ld_dots) + self.ndd

    @property
    def nglob(self):
        return 0

    @staticmethod
    def get_usp(s):
        return 3 + 2 * s

    def get_rho_usps(self):
        usp_set = [self.get_usp(self.slist[n]) for n in range(len(self.slist))]
        usps = []
        for n in range(self.nk0):
            usps.append(usp_set[n])
        for n in range(self.nk1):
            for i in range(3):
                usps.append(usp_set[n] + 1)
        usps.append(SPEC_USPS["grad_rho"])  # for the density gradient
        return usps

    def get_feat_usps(self):
        usp_set = [self.get_usp(self.slist[n]) for n in range(len(self.slist))]
        # -1 index for grad, relies on input checking in __init__
        usp_set.append(3)
        usps = []
        for n in range(self.nk0):
            usps.append(usp_set[n])
        for j, k in self.l1_dots:
            if j < 0:
                assert j == -1
                j = len(usp_set) + j
            if k < 0:
                assert k == -1
                k = len(usp_set) + k
            usps.append(usp_set[j] + usp_set[k] + 2)
        for j, k in self.ld_dots:
            if j < 0:
                assert j == -1
                j = len(usp_set) + j
            if k < 0:
                assert k == -1
                k = len(usp_set) + k
            usps.append(usp_set[j] + usp_set[k] + 2)
        for n in range(self.ndd):
            usps.append(usp_set[n] + 2)
        return usps

    def ueg_vector(self, rho=1.0):
        kf = (rho * 3 * np.pi**2) ** (1.0 / 3)
        vec = [
            kf ** (3 + 2 * s) / (np.pi**2 * (3 + 2 * s))
            for s in self.slist[: self.nk0]
        ]
        # TODO ndd components actually aren't zero for UEG
        return np.array(
            vec + [0] * (len(self.l1_dots) + len(self.ld_dots) + self.ndd),
            dtype=np.float64,
        )

    def get_reasonable_normalizer(self):
        usps = self.get_feat_usps()
        norms = []
        for i in range(self.nk0):
            rho_pow = -0.5
            exp_pow = 0.5 * (1.5 - usps[i])
            norms.append(
                get_normalizer_from_exponent_params(rho_pow, exp_pow, 1.0, 0.03125)
            )
        for i, (j, k) in enumerate(self.l1_dots):
            rho_pow = -1.0
            if j == -1:
                rho_pow -= 0.5
            if k == -1:
                rho_pow -= 0.5
            exp_pow = -0.5 * (usps[i + self.nk0] + 3 * rho_pow)
            norms.append(
                get_normalizer_from_exponent_params(rho_pow, exp_pow, 1.0, 0.03125)
            )
        start = self.nk0 + len(self.l1_dots)
        for i, (j, k) in enumerate(self.ld_dots):
            rho_pow = -1.0
            if j == -1:
                rho_pow -= 0.5
            if k == -1:
                rho_pow -= 0.5
            exp_pow = -0.5 * (usps[i + start] + 3 * rho_pow)
            norms.append(
                get_normalizer_from_exponent_params(rho_pow, exp_pow, 1.0, 0.03125)
            )
        for i in range(self.ndd):
            rho_pow = -0.5
            exp_pow = 0.5 * (1.5 - usps[i - self.ndd])
            norms.append(
                get_normalizer_from_exponent_params(rho_pow, exp_pow, 1.0, 0.03125)
            )
        return norms


class SDMXBaseSettings(BaseSettings):
    """
    Base object for the SDMX-like settings.
    """

    @property
    def ndterms(self):
        return 0

    @property
    def n1terms(self):
        return 0

    @property
    def integral_type(self):
        if hasattr(self, "_integral_type"):
            return self._integral_type
        else:
            return "gauss_diff"


class SADMSettings(SDMXBaseSettings):
    """
    Settings for Spherically Averaged Density Matrix descriptor, i.e.
    the self-repulsion of the exchange hole constructed only from
    the spherically averaged density matrix at a point.

    WARNING: DEPRECATED, especially 'exact' which is not numerically
    accurate. Use SDMXSettings and its variants instead.
    """

    def __init__(self, mode):
        if mode not in ["exact", "smooth"]:
            raise ValueError("Mode must be exact or smooth, got {}".format(mode))
        self.mode = mode

    @property
    def ueg_const(self):
        if self.mode == "smooth":
            return -0.8751999726949563
        else:
            return LDA_FACTOR

    @property
    def nfeat(self):
        return 1

    @staticmethod
    def get_rho_usps():
        return [4]

    def get_feat_usps(self):
        return self.get_rho_usps()

    def ueg_vector(self, rho=1.0):
        return [self.ueg_const * rho ** (4.0 / 3)]

    @property
    def nglob(self):
        return 0

    def get_reasonable_normalizer(self):
        return [DensityNormalizer(1.0 / self.ueg_const, power=(-4.0 / 3))]


class SDMXSettings(SADMSettings):
    def __init__(self, pows):
        """
        Initialize SDMX settings.

        Args:
            pows (list of int): for each number n in pows,
                int dR R^{2-n} rho_smooth(R) is computed. Technically,
                n can be any float, but it should be 0, 1, or 2 for
                normalizability and ability to compute UEG values.
                If not 0, 1, or 2, UEG limit cannot be computed,
                and features might be poorly defined/numerically inaccurate.
        """
        super(SDMXSettings, self).__init__("smooth")
        # if mode not in ['diffse', 'r2', 'r4']:
        #    raise ValueError('Mode must be exact or smooth, got {}'.format(mode))
        self.pows = pows

    @property
    def ueg_const(self):
        consts = []
        for pow in self.pows:
            if pow == 0:
                consts.append(-0.5831374308696952)
            elif pow == 1:
                consts.append(-0.875199972960095)
            elif pow == 2:
                consts.append(-2.073307129927323)
            else:
                raise NotImplementedError("Only have UEG for pow=0,1,2")
        return consts

    @property
    def nfeat(self):
        return len(self.pows)

    def get_feat_usps(self):
        return [3 + n for n in self.pows]

    def ueg_vector(self, rho=1.0):
        return [u * rho ** (1 + n / 3.0) for u, n in zip(self.ueg_const, self.pows)]

    @property
    def nglob(self):
        return 0

    def get_reasonable_normalizer(self):
        return [
            DensityNormalizer(1.0 / u, power=(-1 - n / 3.0))
            for u, n in zip(self.ueg_const, self.pows)
        ]


class SDMXGSettings(SDMXSettings):
    def __init__(self, pows, ndt):
        """
        Initialize SDMXGSettings

        Args:
            pows (list of int): list of 0, 1, 2, see SDMXSettings docstring.
            ndt (int): Number of gradient features H_n^d. Will compute features
                for the first ndt values of n in pows.
        """
        super(SDMXGSettings, self).__init__(pows)
        assert ndt <= len(pows)
        self._ndt = ndt

    @property
    def ndterms(self):
        return self._ndt

    @property
    def ueg_const(self):
        consts = super(SDMXGSettings, self).ueg_const
        for pow in self.pows[: self.ndterms]:
            if pow == 0:
                consts.append(-2.145747704025017)
            elif pow == 1:
                consts.append(-1.7176832947050569)
            elif pow == 2:
                consts.append(-1.526151911492354)
            else:
                raise NotImplementedError("Only have UEG for pow=0,1,2")
        return consts

    @property
    def nfeat(self):
        return len(self.pows) + self.ndterms

    def get_feat_usps(self):
        usps = super(SDMXGSettings, self).get_feat_usps()
        return usps + usps[: self.ndterms]

    def ueg_vector(self, rho=1.0):
        return [
            u * rho ** (1 + n / 3.0)
            for u, n in zip(self.ueg_const, self.pows + self.pows[: self.ndterms])
        ]

    @property
    def nglob(self):
        return 0

    def get_reasonable_normalizer(self):
        return [
            DensityNormalizer(1.0 / u, power=(-1 - n / 3.0))
            for u, n in zip(self.ueg_const, self.pows + self.pows[: self.ndterms])
        ]


class SDMX1Settings(SDMXSettings):
    def __init__(self, pows, n1):
        """
        Initialize SDMX1Settings

        Args:
            pows (list of int): list of 0, 1, 2, see SDMXSettings docstring.
            n1 (int): Number of gradient features H_n^1. Will compute features
                for the first n1 values of n in pows.
        """
        super(SDMX1Settings, self).__init__(pows)
        self.pows = pows
        self._n1 = n1
        assert self._n1 <= len(self.pows)

    @property
    def nfeat(self):
        return len(self.pows) + self._n1

    @property
    def n1terms(self):
        return self._n1

    def get_feat_usps(self):
        res = [3 + n for n in self.pows]
        return res + res[: self._n1]

    def ueg_vector(self, rho=1.0):
        return [u * rho ** (1 + n / 3.0) for u, n in zip(self.ueg_const, self.pows)] + [
            0
        ] * self._n1

    @property
    def nglob(self):
        return 0

    def get_reasonable_normalizer(self):
        ueg_const = self.ueg_const + self.ueg_const[: self._n1]
        pows = self.pows + self.pows[: self._n1]
        return [
            DensityNormalizer(1.0 / u, power=(-1 - n / 3.0))
            for u, n in zip(ueg_const, pows)
        ]


class SDMXG1Settings(SDMXGSettings):
    def __init__(self, pows, nd, n1):
        """
        Initialize SDMXG1Settings

        Args:
            pows (list of int): list of 0, 1, 2, see SDMXSettings docstring.
            nd (int): Number of gradient features H_n^d. Will compute features
                for the first nd values of n in pows.
            n1 (int): Number of gradient features H_n^1. Will compute features
                for the first n1 values of n in pows.
        """
        super(SDMXG1Settings, self).__init__(pows, nd)
        self._n1 = n1

    @property
    def n1terms(self):
        return self._n1

    @property
    def nfeat(self):
        return super(SDMXG1Settings, self).nfeat + self.n1terms

    def get_feat_usps(self):
        res = super(SDMXG1Settings, self).get_feat_usps()
        return res + res[: self._n1]

    def ueg_vector(self, rho=1.0):
        return super(SDMXG1Settings, self).ueg_vector(rho) + [0] * self._n1

    def get_reasonable_normalizer(self):
        norms = super(SDMXG1Settings, self).get_reasonable_normalizer()
        return norms + [
            DensityNormalizer(1.0 / u, power=(-1 - n / 3.0))
            for u, n in zip(self.ueg_const[: self._n1], self.pows[: self._n1])
        ]


class SDMXFullSettings(SDMXBaseSettings):
    mode = "smooth"

    def __init__(self, settings_dict=None):
        """
        Args:
            settings_dict (dict): Each key is float. Each value
                is a 2-tuple. The first item is a list of pows.
                The second item is a 4-list/tuple with the number
                of pows for 0, d, 1, and 1d features.
        """
        self._settings = settings_dict
        try:
            self._check_settings()
        except AssertionError:
            raise ValueError("Invalid settings")

    def _check_settings(self):
        for k, v in self._settings.items():
            assert k >= 1.0
            assert isinstance(k, (int, float))
            assert isinstance(v, (tuple, list))
            assert len(v) == 2
            assert isinstance(v[0], (tuple, list))
            assert isinstance(v[1], (tuple, list))
            assert len(v[1]) == 4
            npow = len(v[0])
            for num in v[1]:
                assert num <= npow

    def _get_num_feat(self, i):
        n = 0
        for k, v in self._settings.items():
            n += v[1][i]
        return n

    @property
    def n0terms(self):
        return self._get_num_feat(0)

    @property
    def ndterms(self):
        return self._get_num_feat(1)

    @property
    def n1terms(self):
        return self._get_num_feat(2)

    @property
    def n1dterms(self):
        return self._get_num_feat(3)

    @property
    def ratios(self):
        return sorted(list(self._settings.keys()))

    def iterate_l0_terms(self, r):
        v0, v1 = self._settings[r]
        for i in range(v1[0]):
            yield v0[i], False
        for i in range(v1[1]):
            yield v0[i], True

    def iterate_l1_terms(self, r):
        v0, v1 = self._settings[r]
        for i in range(v1[2]):
            yield v0[i], False
        for i in range(v1[3]):
            yield v0[i], True

    @property
    def nfeat(self):
        return self.n0terms + self.ndterms + self.n1terms + self.n1dterms

    def _get_ueg_const(self):
        known_ueg_vals = [
            0.5831374308696956,
            0.8751999729599736,
            2.073307129927323,
            2.1457477040250175,
            1.7176832947040594,
            1.526151911492354,
            0.5198776758462206,
            0.8101255945354239,
            1.9940403510297529,
            1.6566509762225055,
            1.3731808970090098,
            1.2642399626720755,
            0.4226571198848177,
            0.7049490157407584,
            1.859127630633614,
            1.0360245380146442,
            0.9154428039042883,
            0.8992798146562154,
        ]
        known_uegs = [
            (1.0, 0, False),
            (1.0, 1, False),
            (1.0, 2, False),
            (1.0, 0, True),
            (1.0, 1, True),
            (1.0, 2, True),
            (1.5, 0, False),
            (1.5, 1, False),
            (1.5, 2, False),
            (1.5, 0, True),
            (1.5, 1, True),
            (1.5, 2, True),
            (2.0, 0, False),
            (2.0, 1, False),
            (2.0, 2, False),
            (2.0, 0, True),
            (2.0, 1, True),
            (2.0, 2, True),
        ]
        known_dict = {k: -1 * v for k, v in zip(known_uegs, known_ueg_vals)}
        return known_dict

    def ueg_vector(self, rho=1.0):
        # (ratio, n, rdr)
        known_dict = self._get_ueg_const()
        for k, v in known_dict.items():
            k2 = (1.0, k[1], k[2])
            known_dict[k] = 0.5 * (known_dict[k] + known_dict[k2])
        uegvec = []
        usps = self.get_feat_usps()
        i = 0
        for ratio in self.ratios:
            for n, rdr in self.iterate_l0_terms(ratio):
                try:
                    uegvec.append(known_dict[ratio, n, rdr] * rho ** (usps[i] / 3.0))
                    i += 1
                except KeyError:
                    raise NotImplementedError("Only support ratio=1,1.5,2, pow=0,1,2")
        for ratio in self.ratios:
            for _ in self.iterate_l1_terms(ratio):
                uegvec.append(0)
        return uegvec

    def get_feat_usps(self):
        usps = []
        for ratio in self.ratios:
            for n, rdr in self.iterate_l0_terms(ratio):
                usps.append(3 + n)
        for ratio in self.ratios:
            for n, rdr in self.iterate_l1_terms(ratio):
                usps.append(3 + n)
        return usps

    def get_reasonable_normalizer(self):
        known_dict = self._get_ueg_const()
        norms = []
        for ratio in self.ratios:
            for n, rdr in self.iterate_l0_terms(ratio):
                try:
                    u = known_dict[ratio, n, rdr]
                except KeyError:
                    raise NotImplementedError("Only support ratio=1,1.5,2, pow=0,1,2")
                norms.append(DensityNormalizer(1.0 / u, power=(-1 - n / 3.0)))
            for n, rdr in self.iterate_l1_terms(ratio):
                try:
                    u = known_dict[ratio, n, rdr]
                except KeyError:
                    raise NotImplementedError("Only support ratio=1,1.5,2, pow=0,1,2")
                norms.append(DensityNormalizer(1.0 / u, power=(-1 - n / 3.0)))
        return norms


class HybridSettings(BaseSettings):
    """
    TODO very rough draft of hybrid settings, to be fully implemented later.
    """

    def __init__(self, alpha, beta, local):
        self.alpha = alpha
        self.beta = beta
        self.local = local

    @property
    def size(self):
        if not self.local:
            return 0
        elif self.beta:  # beta is screened component
            return 2
        else:
            return 1

    @property
    def nglob(self):
        if self.local:
            return 0
        elif self.beta:
            return 2
        else:
            return 1

    @property
    def nfeat(self):
        # TODO
        return 0

    @property
    def get_feat_usps(self):
        raise NotImplementedError

    @property
    def ueg_vector(self, rho=1.0):
        raise NotImplementedError

    def get_reasonable_normalizer(self):
        raise NotImplementedError


ALLOWED_I_SPECS_L0 = ["se", "se_r2", "se_apr2", "se_ap", "se_ap2r2", "se_lapl"]
"""
Allowed spec strings for version i l=0 features.

se: squared-exponential

se_r2: squared-exponential times r^2

se_apr2: squared-exponential times the exponent of the
r' coordinate times r^2

se_ap: squared-exponential times the exponent of the
r' coordinate.

se_ap2r2: squared-exponential times exponent^2 times r^2

se_lapl: laplacian of squared-exponential
"""

ALLOWED_I_SPECS_L1 = ["se_grad", "se_rvec"]
"""
Allowed spec strings for version i l=1 features

se_grad: gradient of squared-exponential

se_rvec: squared-exponential times vector (r'-r)
"""

ALLOWED_J_SPECS = ["se", "se_ar2", "se_a2r4", "se_erf_rinv"]
"""
Allowed spec strings for version j features.
Version k features have the same allowed spec strings

se: squared-exponential

se_ar2: squared-exponential * a * r^2

se_a2r4: squared-exponential * a^2 * r^4

se_erf_rinv: squared-exponential * 1/r with short-range
erf damping
"""
ALLOWED_K_SPECS = ALLOWED_J_SPECS
"""
See ``ALLOWED_J_SPECS``
"""

ALLOWED_RHO_MULTS = ["one", "taumix", "dampmix", "expnt"]
"""
TODO docs here
"""

ALLOWED_RHO_DAMPS = ["none", "exponential", "asymptotic_const"]
"""
TODO docs here
"""

SPEC_USPS = {
    "se": 0,
    "se_r2": -2,
    "se_ar2": 0,
    "se_a2r4": 0,
    "se_erf_rinv": 0,
    "se_ap": 2,
    "se_apr2": 0,
    "se_ap2r2": 2,
    "se_lapl": 2,
    "se_grad": 1,
    "se_rvec": -1,
    "grad_rho": 4,
}
"""
Uniform-scaling powers (USPs) descibe how features scale as the
density is scaled by n_lambda(r) = lambda^3 n(lambda r). If the
USP of a functional F is u, then
F[n_lambda](r) = lambda^u F[n](lambda r)
"""
RHO_MULT_USPS = {
    "one": 0,
    "taumix": 0,
    "dampmix": 0,
    "expnt": 2,
}


class NLDFSettings(BaseSettings):
    """
    NLDFSettings contains the settings for the nonlocal density features,
    which form the core of the CIDER framework. This is an abstract class
    since the feature settings depend on the version. Subclasses are available
    for versions i, j, ij, and k features.
    """

    version = None

    def __init__(
        self,
        sl_level,
        theta_params,
        rho_mult,
    ):
        """
        Initialize NLDFSettings.

        Args:
            sl_level (str): "GGA" or "MGGA", the level of semilocal ingredients
                used to construct the length-scale exponent.
            theta_params (np.ndarray(2 or 3)):
                Settings for the squared-exponential kernel exponent for the r'
                (integrated) coordinate of the features. For version 'k', this
                'exponent' is not used in the squared-exponential but rather
                within the density damping scheme (see rho_damp).
                Should be an array of 3 floats [a0, grad_mul, tau_mul].
                tau_mul is ignored if sl_level="GGA" and may therefore be excluded.
            rho_mult (str): Multiply the density that gets integrated
                by a prefactor. Options: None/'one', 'taumix', 'dampmix', 'expnt'
        """
        self._sl_level = sl_level
        self.theta_params = theta_params
        self.rho_mult = rho_mult
        self.l0_feat_specs = []
        self.l1_feat_specs = []
        self.l1_feat_dots = []
        self.feat_params = []
        self._check_params(self.theta_params)
        if self.sl_level not in ["GGA", "MGGA"]:
            raise ValueError("sl_level must be GGA or MGGA")
        if self.rho_mult not in ALLOWED_RHO_MULTS:
            raise ValueError("Unsupported rho_mult")

    @property
    def sl_level(self):
        # backward compatibility
        if hasattr(self, "_sl_level"):
            return self._sl_level
        else:
            return "MGGA"

    def _check_params(self, params, spec="se"):
        try:
            assert isinstance(params, list)
            for num in params:
                assert isinstance(num, (float, int))
            assert params[0] > 0
            assert params[1] >= 0
            if self.sl_level == "MGGA":
                assert params[2] >= 0
            n = 3 if self.sl_level == "MGGA" else 2
            if spec == "se_erf_rinv":
                assert len(params) == n + 1
            else:
                assert len(params) == n
        except AssertionError:
            raise ValueError("Unsupported feature params")

    @staticmethod
    def _check_specs(specs, allowed):
        for spec in specs:
            if spec not in allowed:
                raise ValueError("Unsupported feature specs")

    @property
    @abstractmethod
    def num_vi_feats(self):
        """Return the number of version i-type features."""

    @property
    @abstractmethod
    def num_feat_param_sets(self):
        """Returns: Number of unique feature parameter sets"""

    @abstractmethod
    def get_feat_usps(self):
        """Returns a list containing the usp of each contracted feature."""

    @property
    @abstractmethod
    def feat_spec_list(self):
        """Return a list containing feature specs in appropriate order."""

    @property
    def nfeat(self):
        """Number of contracted, scale-invariant features."""
        return self.num_feat_param_sets + self.num_vi_feats

    @property
    def nldf_type(self):
        """Same as version, represents type of NLDFs being computed."""
        return self.version

    def _ueg_rho_mult(self, rho):
        if self.rho_mult == "one":
            rho_mult = 1
        elif self.rho_mult == "expnt":
            rho_mult = _get_ueg_expnt(self.theta_params[0], self.theta_params[2], rho)
        else:
            raise NotImplementedError
        return rho_mult

    @abstractmethod
    def get_reasonable_normalizer(self):
        pass


class NLDFSettingsVI(NLDFSettings):

    version = "i"

    def __init__(
        self,
        sl_level,
        theta_params,
        rho_mult,
        l0_feat_specs,
        l1_feat_specs,
        l1_feat_dots,
    ):
        """
        Initialize NLDFSettingsVI object

        Args:
            sl_level (str): "GGA" or "MGGA", the level of semilocal ingredients
                used to construct the length-scale exponent.
            theta_params (np.ndarray):
                Settings for the squared-exponential kernel exponent for the r'
                (integrated) coordinate of the features. For version 'k', this
                'exponent' is not used in the squared-exponential but rather
                within the density damping scheme (see rho_damp).
            rho_mult (str): Multiply the density that gets integrated
                by a prefactor. Options: None/'one', 'taumix', 'dampmix', 'expnt'
            l0_feat_specs (list of str): Each item in the list is a str
                specifying the formula to be used for the scalar (l=0)
                features. See ALLOWED_I_SPECS_L0 for allowed values.
            l1_feat_specs (list of str): Each item in the list is a str
                specifying the formula for the vector (l=1) features.
                See ALLOWED_I_SPECS_L1 for allowed values.
            l1_feat_dots (list of (int, int)): The vector features
                must be contracted with each other (via dot product) to
                form scalar features for the ML model. Each item i in the
                list is a 2-tuple with indexes j,k for features to contract.
                -1 refers to the semilocal density gradient.
                l1feat_ig = einsum('xg,xg->g', l1ints_jg, l1ints_kg)
        """
        super(NLDFSettingsVI, self).__init__(sl_level, theta_params, rho_mult)
        self.l0_feat_specs = l0_feat_specs
        self.l1_feat_specs = l1_feat_specs
        self.l1_feat_dots = l1_feat_dots
        self._check_specs(self.l0_feat_specs, ALLOWED_I_SPECS_L0)
        self._check_specs(self.l1_feat_specs, ALLOWED_I_SPECS_L1)
        _check_l1_dots(self.l1_feat_dots, len(self.l1_feat_specs))

    @property
    def num_feat_param_sets(self):
        return 0

    @property
    def num_vi_feats(self):
        return len(self.l0_feat_specs) + len(self.l1_feat_dots)

    @property
    def feat_spec_list(self):
        return self.l0_feat_specs + self.l1_feat_specs

    def get_feat_usps(self):
        usp0 = RHO_MULT_USPS[self.rho_mult]
        usps = []
        for spec in self.l0_feat_specs:
            usps.append(usp0 + SPEC_USPS[spec])
        for (
            j,
            k,
        ) in self.l1_feat_dots:
            spec1 = "grad_rho" if j == -1 else self.l1_feat_specs[j]
            spec2 = "grad_rho" if k == -1 else self.l1_feat_specs[k]
            usps.append(usp0 + SPEC_USPS[spec1] + SPEC_USPS[spec2])
        return usps

    def ueg_vector(self, rho=1.0):
        l0ueg = []
        l1ueg = [0] * len(self.l1_feat_dots)
        a0 = self.theta_params[0]
        if self.sl_level == "MGGA":
            t0 = self.theta_params[2]
        else:
            t0 = self.theta_params[1]
        rho_mult = self._ueg_rho_mult(rho)
        for spec in self.l0_feat_specs:
            # 'se', 'se_r2', 'se_apr2', 'se_ap', 'se_ap2r2', 'se_lapl'
            expnt = _get_ueg_expnt(a0, t0, rho)
            integral = (np.pi / expnt) ** 1.5
            if spec == "se":
                integral *= 1
            elif spec == "se_r2":
                integral *= 1.5 / expnt
            elif spec == "se_apr2":
                integral *= 1.5
            elif spec == "se_ap":
                integral *= expnt
            elif spec == "se_ap2r2":
                integral *= 1.5 * expnt
            elif spec == "se_lapl":
                integral *= 4 * expnt
            else:
                raise ValueError
            l0ueg.append(rho * rho_mult * integral)
        return np.append(l0ueg, l1ueg).astype(np.float64)

    def get_reasonable_normalizer(self):
        nvi = self.num_vi_feats
        usps = self.get_feat_usps()[self.nfeat - nvi :]
        uegs = self.ueg_vector()[self.nfeat - nvi :]
        a0 = self.theta_params[0]
        if self.sl_level == "MGGA":
            tau_mul = self.theta_params[2]
        else:
            tau_mul = self.theta_params[1]
        norms = []
        for i in range(nvi):
            usp = usps[i]
            ueg = uegs[i]
            if usp == 0 and ueg != 0:
                norms.append(ConstantNormalizer(2.0 / ueg))
            elif usp == 0:
                norms.append(None)
            elif usp == -2:
                norms.append(get_normalizer_from_exponent_params(0.0, 1.0, a0, tau_mul))
            elif usp == 2:
                norms.append(
                    get_normalizer_from_exponent_params(0.0, -1.0, a0, tau_mul)
                )
            elif usp == 5:
                norms.append(
                    get_normalizer_from_exponent_params(-1.0, -1.0, a0, tau_mul)
                )
            else:
                raise NotImplementedError
        return norms


class NLDFSettingsVJ(NLDFSettings):

    version = "j"

    def __init__(
        self,
        sl_level,
        theta_params,
        rho_mult,
        feat_specs,
        feat_params,
    ):
        """
        Initialize NLDFSettingsVJ

        Args:
            sl_level (str): "GGA" or "MGGA", the level of semilocal ingredients
                used to construct the length-scale exponent.
            theta_params (np.ndarray):
                Settings for the squared-exponential kernel exponent for the r'
                (integrated) coordinate of the features. For version 'k', this
                'exponent' is not used in the squared-exponential but rather
                within the density damping scheme (see rho_damp).
            rho_mult (str): Multiply the density that gets integrated
                by a prefactor. Options: None/'one', 'taumix', 'dampmix', 'expnt'
            feat_specs (list of str):
                Each item in the list is a string specifying the formula
                to be used for a feature (see ALLOWED_J_SPECS for options).
                feat_specs[i] uses the parameterization of feat_params[i].
                feat_specs and feat_params must be the same length.
            feat_params (list of np.ndarray):
                Each item in the list is an array with the parameters for the
                feature corresponding to the feat_specs above. Typically, each array
                has three numbers [a0, grad_mul, tau_mul], except erf_rinv which
                has an additional parameter erf_mul for the ratio of the
                erf / rinv exponent to the squared-exponential exponent.
                tau_mul is ignored if sl_level="GGA" and may therefore be excluded.
        """
        super(NLDFSettingsVJ, self).__init__(sl_level, theta_params, rho_mult)
        self.feat_params = feat_params
        self.feat_specs = feat_specs
        self._check_specs(self.feat_specs, ALLOWED_J_SPECS)
        if len(self.feat_params) != len(self.feat_specs):
            raise ValueError("specs and params must have same length")
        for s, p in zip(self.feat_specs, self.feat_params):
            self._check_params(p, spec=s)

    @property
    def num_vi_feats(self):
        return 0

    @property
    def feat_spec_list(self):
        return self.feat_specs

    @property
    def num_feat_param_sets(self):
        return len(self.feat_params)

    def get_feat_usps(self):
        usp0 = RHO_MULT_USPS[self.rho_mult]
        usps = []
        for spec in self.feat_specs:
            usps.append(usp0 + SPEC_USPS[spec])
        return usps

    def ueg_vector(self, rho=1.0):
        ueg_feats = []
        a0t = self.theta_params[0]
        if self.sl_level == "MGGA":
            t0t = self.theta_params[2]
        else:
            t0t = self.theta_params[1]
        rho_mult = self._ueg_rho_mult(rho)
        for spec, params in zip(self.feat_specs, self.feat_params):
            a0i = params[0]
            if self.sl_level == "MGGA":
                t0i = params[2]
            else:
                t0i = params[1]
            a0 = a0i + a0t
            t0 = t0i + t0t
            expnt = _get_ueg_expnt(a0, t0, rho)
            expnt2 = _get_ueg_expnt(a0i, t0i, rho)
            integral = (np.pi / expnt) ** 1.5
            if spec == "se":
                integral *= 1
            elif spec == "se_ar2":
                integral *= 1.5 * expnt2 / expnt
            elif spec == "se_a2r4":
                integral *= 3.75 * (expnt2 / expnt) ** 2
            elif spec == "se_erf_rinv":
                expnt3 = expnt2 * params[-1]
                integral *= np.sqrt(expnt / (expnt + expnt3))
            else:
                raise ValueError
            ueg_feats.append(rho * rho_mult * integral)
        return np.asarray(ueg_feats, dtype=np.float64)

    def get_reasonable_normalizer(self):
        nvj = self.num_feat_param_sets
        uegs = self.ueg_vector()[:nvj]
        usps = self.get_feat_usps()[:nvj]
        a0 = self.theta_params[0]
        if self.sl_level == "MGGA":
            tau_mul = self.theta_params[2]
        else:
            tau_mul = self.theta_params[1]
        norms = []
        for i in range(nvj):
            if usps[i] == 0:
                norms.append(ConstantNormalizer(2.0 / uegs[i]))
            else:
                norms.append(
                    get_normalizer_from_exponent_params(
                        0.0, -0.5 * usps[i], a0, tau_mul
                    )
                )
        return norms


class NLDFSettingsVIJ(NLDFSettings):
    """
    Note:
        When storing features in array, j should go first, followed by i.
    """

    version = "ij"

    def __init__(
        self,
        sl_level,
        theta_params,
        rho_mult,
        l0_feat_specs_i,
        l1_feat_specs_i,
        l1_feat_dots_i,
        feat_specs_j,
        feat_params_j,
    ):
        """
        Initialize NLDFSettingsVIJ object

        Args:
            sl_level (str): "GGA" or "MGGA", the level of semilocal ingredients
                used to construct the length-scale exponent.
            theta_params (np.ndarray):
                Settings for the squared-exponential kernel exponent for the r'
                (integrated) coordinate of the features. For version 'k', this
                'exponent' is not used in the squared-exponential but rather
                within the density damping scheme (see rho_damp).
            rho_mult (str): Multiply the density that gets integrated
                by a prefactor. Options: None/'one', 'taumix', 'dampmix', 'expnt'
            l0_feat_specs_i (list of str): Each item in the list is a str
                specifying the formula to be used for the scalar (l=0)
                features. See ALLOWED_I_SPECS_L0 for allowed values.
            l1_feat_specs_i (list of str): Each item in the list is a str
                specifying the formula for the vector (l=1) features.
                See ALLOWED_I_SPECS_L1 for allowed values.
            l1_feat_dots_i (list of (int, int)): The vector features
                must be contracted with each other (via dot product) to
                form scalar features for the ML model. Each item i in the
                list is a 2-tuple with indexes j,k for features to contract.
                -1 refers to the semilocal density gradient.
                l1feat_ig = einsum('xg,xg->g', l1ints_jg, l1ints_kg)
            feat_specs_j (list of str):
                Each item in the list is a string specifying the formula
                to be used for a feature (see ALLOWED_J_SPECS for options).
                feat_specs[i] uses the parameterization of feat_params[i].
                feat_specs and feat_params must be the same length.
            feat_params_j (list of np.ndarray):
                Each item in the list is an array with the parameters for the
                feature corresponding to the feat_specs above. Typically, each array
                has three numbers [a0, grad_mul, tau_mul], except erf_rinv which
                has an additional parameter erf_mul for the ratio of the
                erf / rinv exponent to the squared-exponential exponent.
                tau_mul is ignored if sl_level="GGA" and may therefore be excluded.
        """
        super(NLDFSettingsVIJ, self).__init__(sl_level, theta_params, rho_mult)
        self.l0_feat_specs = l0_feat_specs_i
        self.l1_feat_specs = l1_feat_specs_i
        self.l1_feat_dots = l1_feat_dots_i
        self.feat_specs = feat_specs_j
        self.feat_params = feat_params_j
        self._check_specs(self.l0_feat_specs, ALLOWED_I_SPECS_L0)
        self._check_specs(self.l1_feat_specs, ALLOWED_I_SPECS_L1)
        self._check_specs(self.feat_specs, ALLOWED_J_SPECS)
        for dot in self.l1_feat_dots:
            if not isinstance(dot, (tuple, list)) or len(dot) != 2:
                raise ValueError("l1_feat_dots items must be 2-tuples")
            for d in dot:
                if not (-1 <= d < len(self.l1_feat_specs)):
                    raise ValueError("Illegal index in feat_dots")
        if len(self.feat_params) != len(self.feat_specs):
            raise ValueError("specs and params must have same length")
        for s, p in zip(self.feat_specs, self.feat_params):
            self._check_params(p, spec=s)

    @property
    def num_feat_param_sets(self):
        return len(self.feat_params)

    @property
    def feat_spec_list(self):
        return self.feat_specs + self.l0_feat_specs + self.l1_feat_specs

    @property
    def num_vi_feats(self):
        return len(self.l0_feat_specs) + len(self.l1_feat_dots)

    def get_feat_usps(self):
        usp0 = RHO_MULT_USPS[self.rho_mult]
        usps = []
        for spec in self.feat_specs:
            usps.append(usp0 + SPEC_USPS[spec])
        for spec in self.l0_feat_specs:
            usps.append(usp0 + SPEC_USPS[spec])
        for (
            j,
            k,
        ) in self.l1_feat_dots:
            spec1 = "grad_rho" if j == -1 else self.l1_feat_specs[j]
            spec2 = "grad_rho" if k == -1 else self.l1_feat_specs[k]
            usps.append(usp0 + SPEC_USPS[spec1] + SPEC_USPS[spec2])
        return usps

    def ueg_vector(self, rho=1.0):
        # a little hacky, but easiest way to avoid
        # redundant code right now
        return np.append(
            NLDFSettingsVJ.ueg_vector(self, rho=rho),
            NLDFSettingsVI.ueg_vector(self, rho=rho),
        )

    def get_reasonable_normalizer(self):
        return NLDFSettingsVJ.get_reasonable_normalizer(
            self
        ) + NLDFSettingsVI.get_reasonable_normalizer(self)


class NLDFSettingsVK(NLDFSettings):

    version = "k"

    def __init__(
        self,
        sl_level,
        theta_params,
        rho_mult,
        feat_params,
        rho_damp,
    ):
        """
        Initialize NLDFSettingsVK

        Args:
            sl_level (str): "GGA" or "MGGA", the level of semilocal ingredients
                used to construct the length-scale exponent.
            theta_params (np.ndarray):
                Settings for the squared-exponential kernel exponent for the r'
                (integrated) coordinate of the features. For version 'k', this
                'exponent' is not used in the squared-exponential but rather
                within the density damping scheme (see rho_damp).
            rho_mult (str): Multiply the density that gets integrated
                by a prefactor. Options: None/'one', 'taumix', 'dampmix', 'expnt'
            feat_params (list of np.ndarray):
                Each item in the list is an array with the parameters for the
                feature corresponding to the feat_specs above. Typically, each array
                has three numbers [a0, grad_mul, tau_mul], except erf_rinv which
                has an additional parameter erf_mul for the ratio of the
                erf / rinv exponent to the squared-exponential exponent.
                tau_mul is ignored if sl_level="GGA" and may therefore be excluded.
            rho_damp (str): Specifies the damping for the
                density. Options: 'none', 'exponential', 'asymptotic_const'
        """
        super(NLDFSettingsVK, self).__init__(sl_level, theta_params, rho_mult)
        self.feat_specs = ["se"] * len(feat_params)
        self.feat_params = feat_params
        self._check_specs(self.feat_specs, ALLOWED_K_SPECS)
        for s, p in zip(self.feat_specs, self.feat_params):
            self._check_params(p, spec=s)
        self.rho_damp = rho_damp

    @property
    def num_feat_param_sets(self):
        return len(self.feat_params)

    @property
    def feat_spec_list(self):
        return self.feat_specs

    @property
    def num_vi_feats(self):
        return 0

    def get_feat_usps(self):
        usp0 = RHO_MULT_USPS[self.rho_mult]
        return [usp0 + SPEC_USPS[spec] for spec in self.feat_specs]

    def ueg_vector(self, rho=1.0):
        a0t = self.theta_params[0]
        if self.sl_level == "MGGA":
            t0t = self.theta_params[2]
        else:
            t0t = self.theta_params[1]
        rho_mult = self._ueg_rho_mult(rho)
        expnt_theta = _get_ueg_expnt(a0t, t0t, rho)
        ueg_feats = []
        for spec, params in zip(self.feat_specs, self.feat_params):
            if self.sl_level == "MGGA":
                t0i = params[2]
            else:
                t0i = params[1]
            expnt = _get_ueg_expnt(params[0], t0i, rho)
            # integral part
            integral = (np.pi / expnt) ** 1.5
            # damping part
            integral *= np.exp(-1.5 * expnt_theta / expnt)
            # multipliers for more complicated integrals
            if spec == "se":
                integral *= 1
            elif spec == "se_ar2":
                integral *= 1.5
            elif spec == "se_a2r4":
                integral *= 3.75
            elif spec == "se_erf_rinv":
                expnt3 = expnt * params[-1]
                integral *= np.sqrt(expnt / (expnt + expnt3))
            else:
                raise ValueError
            ueg_feats.append(rho * rho_mult * integral)
        return np.asarray(ueg_feats, dtype=np.float64)

    def get_reasonable_normalizer(self):
        nvk = self.num_feat_param_sets
        uegs = self.ueg_vector()
        usps = self.get_feat_usps()
        norms = []
        for i in range(nvk):
            if usps[i] == 0:
                norms.append(ConstantNormalizer(2.0 / uegs[i]))
            else:
                a0 = self.theta_params[0]
                if self.sl_level == "MGGA":
                    tau_mul = self.theta_params[2]
                else:
                    tau_mul = self.theta_params[1]
                norms.append(
                    get_normalizer_from_exponent_params(
                        0.0, -0.5 * usps[i], a0, tau_mul
                    )
                )
        return norms


class SemilocalSettings(BaseSettings):
    """
    Semilocal feature set. Currently only supports meta-GGA.
    Should not be edited from default in general.
    """

    def __init__(self, mode="nst"):
        # mode should be nst or npa (MGGA) or ns or np (GGA).
        self.level = None
        if mode in ["nst", "npa"]:
            self.level = "MGGA"
        elif mode in ["ns", "np"]:
            self.level = "GGA"
        else:
            raise ValueError("Mode must be nst, npa, ns, or np.")
        self.mode = mode

    @property
    def nfeat(self):
        if self.level == "MGGA":
            return 3
        else:
            return 2

    def get_feat_usps(self):
        if self.mode == "nst":
            return [3, 8, 5]
        elif self.mode == "npa":
            return [3, 0, 0]
        elif self.mode == "ns":
            return [3, 8]
        else:
            return [3, 0]

    def ueg_vector(self, rho=1.0):
        if self.mode == "nst":
            return np.array([rho, 0.0, CFC * rho ** (5.0 / 3)])
        elif self.mode == "npa":
            return np.array([rho, 0.0, 1.0])
        else:
            return np.array([rho, 0.0])

    def get_reasonable_normalizer(self):
        return [None] * self.nfeat


class FeatureSettings(BaseSettings):
    """
    The FeatureSettings object is a container for the settings of each
    type of feature in the model.
    """

    def __init__(
        self,
        sl_settings=None,
        nldf_settings=None,
        nlof_settings=None,
        sdmx_settings=None,
        hyb_settings=None,
        normalizers=None,
    ):
        """
        Initialize FeatureSettings.

        Args:
            sl_settings (SemilocalSettings or EmptySettings):
                Semilocal feature settings
            nldf_settings (NLDFSettings or EmptySettings):
                Nonlocal density feature settings
            nlof_settings (FracLaplSettings or EmptySettings):
                Nonlocal orbital feature settings
            sdmx_settings (SDMXBaseSettings or EmptySettings):
                Spherically averaged EXX settings
            hyb_settings (HybridSettings or EmptySettings):
                (Local) hybrid DFT settings
            normalizers (FeatNormalizerList): List of normalizers for features
        """
        self.sl_settings = EmptySettings() if sl_settings is None else sl_settings
        self.nldf_settings = EmptySettings() if nldf_settings is None else nldf_settings
        self.nlof_settings = EmptySettings() if nlof_settings is None else nlof_settings
        self.sdmx_settings = EmptySettings() if sdmx_settings is None else sdmx_settings
        self.hyb_settings = EmptySettings() if hyb_settings is None else hyb_settings
        self.normalizers = (
            FeatNormalizerList([None] * self.nfeat)
            if normalizers is None
            else normalizers
        )
        if self.hyb_settings.nfeat != 0:
            raise NotImplementedError("Hybrid DFT")

    @property
    def has_sl(self):
        return not self.sl_settings.is_empty

    @property
    def has_nldf(self):
        return not self.nldf_settings.is_empty

    @property
    def has_nlof(self):
        return not self.nlof_settings.is_empty

    @property
    def has_sdmx(self):
        return not self.sdmx_settings.is_empty

    @property
    def nfeat(self):
        return (
            self.sl_settings.nfeat
            + self.nldf_settings.nfeat
            + self.nlof_settings.nfeat
            + self.sdmx_settings.nfeat
            + self.hyb_settings.nfeat
        )

    def get_feat_loc(self):
        return np.cumsum(
            [
                0,
                self.sl_settings.nfeat,
                self.nldf_settings.nfeat,
                self.nlof_settings.nfeat,
                self.sdmx_settings.nfeat,
                self.hyb_settings.nfeat,
            ]
        )

    def get_feat_loc_dict(self):
        loc = self.get_feat_loc()
        return {
            "sl": loc[0],
            "nldf": loc[1],
            "nlof": loc[2],
            "sadm": loc[3],
            "hyb": loc[4],
            "end": loc[5],
        }

    def get_feat_usps(self, with_normalizers=False):
        res = np.concatenate(
            [
                self.sl_settings.get_feat_usps(),
                self.nldf_settings.get_feat_usps(),
                self.nlof_settings.get_feat_usps(),
                self.sdmx_settings.get_feat_usps(),
                self.hyb_settings.get_feat_usps(),
            ]
        )
        if with_normalizers:
            res += np.array(self.normalizers.get_usps())
        return res

    def ueg_vector(self, rho=1.0, with_normalizers=False):
        res = np.concatenate(
            [
                self.sl_settings.ueg_vector(rho),
                self.nldf_settings.ueg_vector(rho),
                self.nlof_settings.ueg_vector(rho),
                self.sdmx_settings.ueg_vector(rho),
                self.hyb_settings.ueg_vector(rho),
            ]
        )
        if with_normalizers:
            res *= self.normalizers.ueg_vector(rho)
        return res

    def get_reasonable_normalizer(self):
        return (
            self.sl_settings.get_reasonable_normalizer()
            + self.nldf_settings.get_reasonable_normalizer()
            + self.nlof_settings.get_reasonable_normalizer()
            + self.sdmx_settings.get_reasonable_normalizer()
            + self.hyb_settings.get_reasonable_normalizer()
        )

    def assign_reasonable_normalizer(self):
        self.normalizers = FeatNormalizerList(self.get_reasonable_normalizer())


def dtauw(rho, sigma):
    return -sigma / (8 * rho**2 + 1e-16), 1 / (8 * rho + 1e-16)


def get_uniform_tau(rho):
    return (3.0 / 10) * (3 * np.pi**2) ** (2.0 / 3) * rho ** (5.0 / 3)


def get_single_orbital_tau(rho, mag_grad):
    return mag_grad**2 / (8 * rho + 1e-16)


def get_s2(rho, sigma):
    # TODO should this cutoff not be needed if everything else is stable?
    # rho = np.maximum(1e-10, rho)
    cond = rho < ALPHA_TOL
    b = 2 * (3 * np.pi * np.pi) ** (1.0 / 3)
    s = np.sqrt(sigma) / (b * rho ** (4.0 / 3) + 1e-16)
    s[cond] = 0.0
    return s * s


def ds2(rho, sigma):
    # s = |nabla n| / (b * n)
    # TODO should this cutoff not be needed if everything else is stable?
    cond = rho < ALPHA_TOL
    rho = np.maximum(ALPHA_TOL, rho)
    b = 2 * (3 * np.pi * np.pi) ** (1.0 / 3)
    s = np.sqrt(sigma) / (b * rho ** (4.0 / 3) + 1e-16)
    s2 = s**2
    res = -8.0 * s2 / (3 * rho + 1e-16), 1 / (b * rho ** (4.0 / 3) + 1e-16) ** 2
    res[0][cond] = 0.0
    res[1][cond] = 0.0
    return res


def get_alpha(rho, sigma, tau):
    cond = rho < ALPHA_TOL
    rho = np.maximum(ALPHA_TOL, rho)
    tau0 = get_uniform_tau(rho)
    tauw = get_single_orbital_tau(rho, np.sqrt(sigma))
    # TODO this numerical stability trick is a bit of a hack.
    # Should make spline support small negative alpha
    # instead, for the sake of clean code and better stability.
    alpha = np.maximum((tau - tauw), 0) / tau0
    alpha[cond] = 0
    return alpha
    # return np.maximum((tau - tauw), 0) / tau0


def dalpha(rho, sigma, tau):
    cond = rho < ALPHA_TOL
    rho = np.maximum(ALPHA_TOL, rho)
    tau0 = get_uniform_tau(rho)
    tauw = sigma / (8 * rho)
    dwdn, dwds = -sigma / (8 * rho * rho), 1 / (8 * rho)
    dadn = 5.0 * (tauw - tau) / (3 * tau0 * rho) - dwdn / tau0
    dadsigma = -dwds / tau0
    dadtau = 1 / tau0
    # cond = (tau - tauw) / tau0 < -0.1
    dadn[cond] = 0
    dadsigma[cond] = 0
    dadtau[cond] = 0
    return dadn, dadsigma, dadtau



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np
from pyscf.dft.gen_grid import Grids
from pyscf.dft.numint import NumInt

from ciderpress.dft.debug_numint import get_get_exponent, get_nonlocal_features


def get_nldf_numint(
    mol, grids, dms, gg_kwargs, vv_gg_kwargs, version="i", inner_grids_level=3
):
    if dms.ndim == 2:
        return _get_nldf_numint(
            mol,
            grids,
            dms,
            gg_kwargs,
            vv_gg_kwargs,
            version=version,
            inner_grids_level=inner_grids_level,
        )[None, :]
    else:
        assert dms.shape[0] == 2
        dms = dms * 2
        res0 = _get_nldf_numint(
            mol,
            grids,
            dms[0],
            gg_kwargs,
            vv_gg_kwargs,
            version=version,
            inner_grids_level=inner_grids_level,
        )
        res1 = _get_nldf_numint(
            mol,
            grids,
            dms[1],
            gg_kwargs,
            vv_gg_kwargs,
            version=version,
            inner_grids_level=inner_grids_level,
        )
        return np.stack([res0, res1])


def _get_nldf_numint(
    mol, grids, dms, gg_kwargs, vv_gg_kwargs, version="i", inner_grids_level=3
):
    ni = NumInt()
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, 1, False)
    max_memory = 2000
    assert nset == 1

    get_exponent_r = get_get_exponent(gg_kwargs)
    get_exponent_rp = get_get_exponent(vv_gg_kwargs)

    assert nset == 1
    smgrids = Grids(mol)
    smgrids.level = inner_grids_level
    smgrids.prune = None
    smgrids.build()

    nfeat = 12 if version == "i" else 4
    desc = np.empty((0, nfeat))
    ao_deriv = 1
    vvrho = np.empty([nset, 5, 0])
    vvweight = np.empty([nset, 0])
    vvcoords = np.empty([nset, 0, 3])
    for ao, mask, weight, coords in ni.block_loop(
        mol, smgrids, nao, ao_deriv, max_memory
    ):
        rhotmp = np.empty([0, 5, weight.size])
        weighttmp = np.empty([0, weight.size])
        coordstmp = np.empty([0, weight.size, 3])
        for idm in range(nset):
            rho = make_rho(idm, ao, mask, "MGGA")
            rho = np.expand_dims(rho, axis=0)
            rhotmp = np.concatenate((rhotmp, rho), axis=0)
            weighttmp = np.concatenate(
                (weighttmp, np.expand_dims(weight, axis=0)),
                axis=0,
            )
            coordstmp = np.concatenate(
                (coordstmp, np.expand_dims(coords, axis=0)),
                axis=0,
            )
        vvrho = np.concatenate((vvrho, rhotmp), axis=2)
        vvweight = np.concatenate((vvweight, weighttmp), axis=1)
        vvcoords = np.concatenate((vvcoords, coordstmp), axis=1)
    aow = None
    for ao, mask, weight, coords in ni.block_loop(
        mol, grids, nao, ao_deriv, max_memory
    ):
        aow = np.ndarray(ao[0].shape, order="F", buffer=aow)
        for idm in range(nset):
            rho = make_rho(idm, ao, mask, "MGGA")
            feat = get_nonlocal_features(
                rho,
                coords,
                vvrho[idm],
                vvweight[idm],
                vvcoords[idm],
                get_exponent_r,
                get_exponent_rp,
                version=version,
            )
            desc = np.append(desc, feat, axis=0)
    return desc



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np
from ase.utils.timing import Timer
from pyscf import lib
from pyscf.dft import numint
from pyscf.dft.gen_grid import ALIGNMENT_UNIT, BLKSIZE, NBINS
from pyscf.dft.numint import _dot_ao_ao_sparse, _scale_ao_sparse, eval_ao

import ciderpress.pyscf.frac_lapl as nlof
from ciderpress.dft.plans import (
    FracLaplPlan,
    SemilocalPlan,
    get_rho_tuple_with_grad_cross,
    vxc_tuple_to_array,
)
from ciderpress.dft.settings import FeatureSettings
from ciderpress.dft.xc_evaluator import MappedXC
from ciderpress.dft.xc_evaluator2 import MappedXC2
from ciderpress.pyscf.nldf_convolutions import PyscfNLDFGenerator
from ciderpress.pyscf.sdmx import EXXSphGenerator

DEFAULT_RHOCUT = 1e-9


def _tau_dot_sparse(
    bra, ket, wv, nbins, screen_index, pair_mask, ao_loc, out=None, aow=None
):
    """Similar to _tau_dot, while sparsity is explicitly considered. Note the
    return may have ~1e-13 difference to _tau_dot.
    """
    nao = bra.shape[1]
    if out is None:
        out = np.zeros((nao, nao), dtype=bra.dtype)
    hermi = 1
    aow = _scale_ao_sparse(ket[1], wv, screen_index, ao_loc, out=aow)
    _dot_ao_ao_sparse(
        bra[1], aow, None, nbins, screen_index, pair_mask, ao_loc, hermi, out
    )
    aow = _scale_ao_sparse(ket[2], wv, screen_index, ao_loc, out=aow)
    _dot_ao_ao_sparse(
        bra[2], aow, None, nbins, screen_index, pair_mask, ao_loc, hermi, out
    )
    aow = _scale_ao_sparse(ket[3], wv, screen_index, ao_loc, out=aow)
    _dot_ao_ao_sparse(
        bra[3], aow, None, nbins, screen_index, pair_mask, ao_loc, hermi, out
    )
    return out


def _get_sdmx_orbs(ni, mol, coords, ao):
    ni.timer.start("sdmx ao")
    if ni.has_sdmx and ni.sdmxgen.fast:
        if ao is None:
            sdmx_ao = eval_ao(mol, coords, deriv=0)
        else:
            sdmx_ao = ao[0] if ao.ndim == 3 else ao
        sdmx_cao = ni.sdmxgen.get_cao(mol, coords, save_buf=True)
    elif ni.has_sdmx:
        sdmx_ao, sdmx_cao = ni.sdmxgen.get_orb_vals(mol, coords, save_buf=True)
    else:
        sdmx_ao, sdmx_cao = None, None
    ni.timer.stop("sdmx ao")
    return sdmx_ao, sdmx_cao


def nr_rks(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    """

    Args:
        ni (CiderNumInt):
        mol (pyscf.gto.Mole):
        grids (pyscf.dft.gen_grid.Grids):
        xc_code:
        dms:
        relativity:
        hermi:
        max_memory:
        verbose:

    Returns:

    """
    ni.timer.start("nr_rks")
    if relativity != 0:
        raise NotImplementedError

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 1)

    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()
    cutoff = grids.cutoff * 1e2
    nbins = NBINS * 2 - int(NBINS * np.log(cutoff) / np.log(grids.cutoff))

    if dms.ndim == 2:
        dms = dms[None, :, :]

    nelec = np.zeros(nset)
    excsum = np.zeros(nset)
    vmat = np.zeros((nset, nao, nao))

    def block_loop(ao_deriv):
        if ni.has_sdmx:
            extra_ao = ni.sdmxgen.get_extra_ao(mol)
        else:
            extra_ao = 0
        for ao, mask, weight, coords in ni.block_loop(
            mol, grids, nao, ao_deriv, max_memory=max_memory, extra_ao=extra_ao
        ):
            sdmx_ao, sdmx_cao = _get_sdmx_orbs(ni, mol, coords, ao)
            for i in range(nset):
                rho = make_rho(i, ao, mask, xctype)
                ni.timer.start("sdmx fwd")
                if ni.has_sdmx:
                    sdmx_feat = ni.sdmxgen.get_features(
                        dms[i],
                        mol,
                        coords,
                        ao=sdmx_ao,
                        cao=sdmx_cao,
                    )
                else:
                    sdmx_feat = None
                ni.timer.stop("sdmx fwd")
                ni.timer.start("xc cider")
                exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                    xc_code, rho, None, sdmx_feat, deriv=1, xctype=xctype
                )[:2]
                ni.timer.stop("xc cider")
                assert vxc_nldf is None
                ni.timer.start("sdmx bwd")
                if ni.has_sdmx:
                    ni.sdmxgen.get_vxc_(vmat[i], vxc_sdmx[0] * weight)
                ni.timer.stop("sdmx bwd")
                den = rho[0] * weight
                nelec[i] += den.sum()
                excsum[i] += np.dot(den, exc)
                wv = weight * vxc
                yield i, ao, mask, wv
        # if ni.has_sdmx:
        #     ni.sdmxgen.reset_buffers()

    buffers = None
    pair_mask = mol.get_overlap_cond() < -np.log(ni.cutoff)
    if any(x in xc_code.upper() for x in ("CC06", "CS", "BR89", "MK00")):
        raise NotImplementedError("laplacian in meta-GGA method")
    ao_deriv = 1
    v1 = np.zeros_like(vmat)
    for i, ao, mask, wv in block_loop(ao_deriv):
        vmats, buffers = ni.contract_wv(
            ao,
            wv,
            nbins,
            mask,
            pair_mask,
            ao_loc,
            vmats=(vmat[i], v1[i]),
            buffers=buffers,
        )
    vmat = lib.hermi_sum(vmat, axes=(0, 2, 1))
    vmat += v1

    if ni.has_sdmx:
        ni.sdmxgen._cached_ao_data = None

    if nset == 1:
        nelec = nelec[0]
        excsum = excsum[0]
        vmat = vmat[0]
    if isinstance(dms, np.ndarray):
        dtype = dms.dtype
    else:
        dtype = np.result_type(*dms)
    if vmat.dtype != dtype:
        vmat = np.asarray(vmat, dtype=dtype)
    # if ni.has_sdmx:
    #    ni.sdmxgen.reset_buffers()
    ni.timer.stop("nr_rks")
    return nelec, excsum, vmat


def nr_uks(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    ni.timer.start("nr_uks")
    if relativity != 0:
        raise NotImplementedError

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 2)

    dma, dmb = numint._format_uks_dm(dms)
    if dma.ndim == 2:
        dma = dma[None, :, :]
        dmb = dmb[None, :, :]
        is_2d = True
    else:
        is_2d = False
    nao = dma.shape[-1]
    make_rhoa, nset = ni._gen_rho_evaluator(mol, dma, hermi, False, grids)[:2]
    make_rhob = ni._gen_rho_evaluator(mol, dmb, hermi, False, grids)[0]
    ao_loc = mol.ao_loc_nr()
    cutoff = grids.cutoff * 1e2
    nbins = NBINS * 2 - int(NBINS * np.log(cutoff) / np.log(grids.cutoff))
    if ni.has_sdmx:
        dm_for_sdmx = [np.stack([dma[i], dmb[i]]) for i in range(nset)]
    else:
        dm_for_sdmx = None

    nelec = np.zeros((2, nset))
    excsum = np.zeros(nset)
    vmat = np.zeros((2, nset, nao, nao))

    def block_loop(ao_deriv):
        if ni.has_sdmx:
            extra_ao = ni.sdmxgen.get_extra_ao(mol)
        else:
            extra_ao = 0
        for ao, mask, weight, coords in ni.block_loop(
            mol, grids, nao, ao_deriv, max_memory=max_memory, extra_ao=extra_ao
        ):
            sdmx_ao, sdmx_cao = _get_sdmx_orbs(ni, mol, coords, ao)
            for i in range(nset):
                rho_a = make_rhoa(i, ao, mask, xctype)
                rho_b = make_rhob(i, ao, mask, xctype)
                rho = (rho_a, rho_b)
                ni.timer.start("sdmx fwd")
                if ni.has_sdmx:
                    sdmx_feat = ni.sdmxgen.get_features(
                        dm_for_sdmx[i],
                        mol,
                        coords,
                        ao=sdmx_ao,
                        cao=sdmx_cao,
                    )
                else:
                    sdmx_feat = None
                ni.timer.stop("sdmx fwd")
                ni.timer.start("xc cider")
                exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                    xc_code, rho, None, sdmx_feat, deriv=1, xctype=xctype
                )[:2]
                ni.timer.stop("xc cider")
                assert vxc_nldf is None
                ni.timer.start("sdmx bwd")
                if ni.has_sdmx:
                    ni.sdmxgen.get_vxc_(vmat[:, i], vxc_sdmx * weight)
                ni.timer.stop("sdmx bwd")
                den_a = rho_a[0] * weight
                den_b = rho_b[0] * weight
                nelec[0, i] += den_a.sum()
                nelec[1, i] += den_b.sum()
                excsum[i] += np.dot(den_a, exc)
                excsum[i] += np.dot(den_b, exc)
                wv = weight * vxc
                yield i, ao, mask, wv

    buffers = None
    pair_mask = mol.get_overlap_cond() < -np.log(ni.cutoff)
    if any(x in xc_code.upper() for x in ("CC06", "CS", "BR89", "MK00")):
        raise NotImplementedError("laplacian in meta-GGA method")
    ao_deriv = 1
    v1 = np.zeros_like(vmat)
    for i, ao, mask, wv in block_loop(ao_deriv):
        buffers = ni.contract_wv(
            ao,
            wv[0],
            nbins,
            mask,
            pair_mask,
            ao_loc,
            vmats=(vmat[0, i], v1[0, i]),
            buffers=buffers,
        )[1]
        buffers = ni.contract_wv(
            ao,
            wv[1],
            nbins,
            mask,
            pair_mask,
            ao_loc,
            vmats=(vmat[1, i], v1[1, i]),
            buffers=buffers,
        )[1]
    vmat = lib.hermi_sum(vmat.reshape(-1, nao, nao), axes=(0, 2, 1)).reshape(
        2, nset, nao, nao
    )
    vmat += v1

    if ni.has_sdmx:
        ni.sdmxgen._cached_ao_data = None

    if isinstance(dma, np.ndarray) and is_2d:
        vmat = vmat[:, 0]
        nelec = nelec.reshape(2)
        excsum = excsum[0]

    dtype = np.result_type(dma, dmb)
    if vmat.dtype != dtype:
        vmat = np.asarray(vmat, dtype=dtype)
    ni.timer.stop("nr_uks")
    return nelec, excsum, vmat


def nr_rks_nldf(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    """

    Args:
        ni (NLDFNumInt):
        mol (pyscf.gto.Mole):
        grids (pyscf.dft.gen_grid.Grids):
        xc_code (str):
        dms (numpy.ndarray):
        relativity:
        hermi:
        max_memory:
        verbose:

    Returns:

    """
    if not hasattr(grids, "grids_indexer"):
        raise ValueError("Grids object must have indexer for NLDF evaluation")

    if relativity != 0:
        raise NotImplementedError

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 1)
    assert ni.has_nldf

    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()
    cutoff = grids.cutoff * 1e2
    nbins = NBINS * 2 - int(NBINS * np.log(cutoff) / np.log(grids.cutoff))

    if dms.ndim == 2:
        dms = dms[None, :, :]

    nelec = np.zeros(nset)
    excsum = np.zeros(nset)
    vmat = np.zeros((nset, nao, nao))
    nrho = 5 if xctype == "MGGA" else 4
    if not ni.settings.nlof_settings.is_empty:
        nrho += ni.settings.nlof_settings.nrho
    rho_full = np.zeros((nset, nrho, grids.weights.size), dtype=np.float64, order="C")
    wv_full = np.zeros((nset, nrho, grids.weights.size), dtype=np.float64, order="C")

    num_nldf = ni.settings.nldf_settings.nfeat
    vxc_nldf_full = np.zeros(
        (nset, num_nldf, grids.weights.size), dtype=np.float64, order="C"
    )

    def block_loop(ao_deriv):
        ip0 = 0
        for ao, mask, weight, coords in ni.block_loop(
            mol, grids, nao, ao_deriv, max_memory=max_memory
        ):
            ip1 = ip0 + weight.size
            for i in range(nset):
                yield i, ip0, ip1, ao, mask, weight, coords
            ip0 = ip1

    if any(x in xc_code.upper() for x in ("CC06", "CS", "BR89", "MK00")):
        raise NotImplementedError("laplacian in meta-GGA method")
    ao_deriv = 1
    for i, ip0, ip1, ao, mask, weight, coords in block_loop(ao_deriv):
        rho_full[i, :, ip0:ip1] = make_rho(i, ao, mask, xctype)

    par_atom = False
    extra_ao = ni.nldfgen.get_extra_ao()
    if ni.has_sdmx:
        extra_ao += ni.sdmxgen.get_extra_ao(mol)
    if par_atom:
        raise NotImplementedError
    else:
        nldf_feat = []
        for idm in range(nset):
            nldf_feat.append(ni.nldfgen.get_features(rho_full[idm]))
        ip0 = 0
        for mask, weight, coords in ni.extra_block_loop(
            mol, grids, max_memory=max_memory, extra_ao=extra_ao
        ):
            ip1 = ip0 + weight.size
            sdmx_ao, sdmx_cao = _get_sdmx_orbs(ni, mol, coords, None)
            for idm in range(nset):
                rho = rho_full[idm, :, ip0:ip1]
                if ni.has_sdmx:
                    sdmx_feat = ni.sdmxgen.get_features(
                        dms[idm], mol, coords, ao=sdmx_ao, cao=sdmx_cao
                    )
                else:
                    sdmx_feat = None
                exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                    xc_code,
                    rho,
                    nldf_feat[idm][:, ip0:ip1],
                    sdmx_feat,
                    deriv=1,
                    xctype=xctype,
                )[:2]
                if ni.has_sdmx:
                    ni.sdmxgen.get_vxc_(vmat[idm], vxc_sdmx[0] * weight)
                vxc_nldf_full[idm, :, ip0:ip1] = vxc_nldf * weight
                den = rho[0] * weight
                nelec[idm] += den.sum()
                excsum[idm] += np.dot(den, exc)
                wv_full[idm, :, ip0:ip1] = weight * vxc
            ip0 = ip1
        for idm in range(nset):
            wv_full[idm, :, :] += ni.nldfgen.get_potential(vxc_nldf_full[idm])

    buffers = None
    pair_mask = mol.get_overlap_cond() < -np.log(ni.cutoff)
    v1 = np.zeros_like(vmat)
    for i, ip0, ip1, ao, mask, weight, coords in block_loop(ao_deriv):
        wv = wv_full[i, :, ip0:ip1]
        vmats, buffers = ni.contract_wv(
            ao,
            wv,
            nbins,
            mask,
            pair_mask,
            ao_loc,
            vmats=(vmat[i], v1[i]),
            buffers=buffers,
        )
    vmat = lib.hermi_sum(vmat, axes=(0, 2, 1))
    vmat += v1

    if nset == 1:
        nelec = nelec[0]
        excsum = excsum[0]
        vmat = vmat[0]
    if isinstance(dms, np.ndarray):
        dtype = dms.dtype
    else:
        dtype = np.result_type(*dms)
    if vmat.dtype != dtype:
        vmat = np.asarray(vmat, dtype=dtype)
    return nelec, excsum, vmat


def nr_uks_nldf(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    ni.timer.start("nr_uks")
    if relativity != 0:
        raise NotImplementedError

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 2)

    dma, dmb = numint._format_uks_dm(dms)
    if dma.ndim == 2:
        dma = dma[None, :, :]
        dmb = dmb[None, :, :]
        is_2d = True
    else:
        is_2d = False
    nao = dma.shape[-1]
    make_rhoa, nset = ni._gen_rho_evaluator(mol, dma, hermi, False, grids)[:2]
    make_rhob = ni._gen_rho_evaluator(mol, dmb, hermi, False, grids)[0]
    ao_loc = mol.ao_loc_nr()
    cutoff = grids.cutoff * 1e2
    nbins = NBINS * 2 - int(NBINS * np.log(cutoff) / np.log(grids.cutoff))
    if ni.has_sdmx:
        dm_for_sdmx = [np.stack([dma[i], dmb[i]]) for i in range(nset)]
    else:
        dm_for_sdmx = None

    nelec = np.zeros((2, nset))
    excsum = np.zeros(nset)
    vmat = np.zeros((2, nset, nao, nao))
    nrho = 5 if xctype == "MGGA" else 4
    if not ni.settings.nlof_settings.is_empty:
        nrho += ni.settings.nlof_settings.nrho
    rhoa_full = np.zeros((nset, nrho, grids.weights.size), dtype=np.float64, order="C")
    rhob_full = np.zeros((nset, nrho, grids.weights.size), dtype=np.float64, order="C")
    wva_full = np.zeros((nset, nrho, grids.weights.size), dtype=np.float64, order="C")
    wvb_full = np.zeros((nset, nrho, grids.weights.size), dtype=np.float64, order="C")

    num_nldf = ni.settings.nldf_settings.nfeat
    vxc_nldf_full = np.zeros(
        (nset, 2, num_nldf, grids.weights.size), dtype=np.float64, order="C"
    )

    def block_loop(ao_deriv):
        ip0 = 0
        for ao, mask, weight, coords in ni.block_loop(
            mol, grids, nao, ao_deriv, max_memory=max_memory
        ):
            ip1 = ip0 + weight.size
            for i in range(nset):
                yield i, ip0, ip1, ao, mask, weight, coords
            ip0 = ip1

    if any(x in xc_code.upper() for x in ("CC06", "CS", "BR89", "MK00")):
        raise NotImplementedError("laplacian in meta-GGA method")
    ao_deriv = 1
    for i, ip0, ip1, ao, mask, weight, coords in block_loop(ao_deriv):
        rhoa_full[i, :, ip0:ip1] = make_rhoa(i, ao, mask, xctype)
        rhob_full[i, :, ip0:ip1] = make_rhob(i, ao, mask, xctype)

    par_atom = False
    extra_ao = ni.nldfgen.get_extra_ao()
    if ni.has_sdmx:
        extra_ao += ni.sdmxgen.get_extra_ao(mol)
    if par_atom:
        raise NotImplementedError
    else:
        nldf_feat = []
        for idm in range(nset):
            nldf_feat.append(
                np.stack(
                    [
                        ni.nldfgen.get_features(rhoa_full[idm], spin=0),
                        ni.nldfgen.get_features(rhob_full[idm], spin=1),
                    ]
                )
            )
        ip0 = 0
        for mask, weight, coords in ni.extra_block_loop(
            mol, grids, max_memory=max_memory, extra_ao=extra_ao
        ):
            ip1 = ip0 + weight.size
            sdmx_ao, sdmx_cao = _get_sdmx_orbs(ni, mol, coords, None)
            for idm in range(nset):
                rho_a = rhoa_full[idm, :, ip0:ip1]
                rho_b = rhob_full[idm, :, ip0:ip1]
                rho = (rho_a, rho_b)
                if ni.has_sdmx:
                    sdmx_feat = ni.sdmxgen.get_features(
                        dm_for_sdmx[i],
                        mol,
                        coords,
                        ao=sdmx_ao,
                        cao=sdmx_cao,
                    )
                else:
                    sdmx_feat = None
                exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                    xc_code,
                    rho,
                    nldf_feat[idm][..., ip0:ip1],
                    sdmx_feat,
                    deriv=1,
                    xctype=xctype,
                )[:2]
                if ni.has_sdmx:
                    ni.sdmxgen.get_vxc_(vmat[:, i], vxc_sdmx * weight)
                vxc_nldf_full[idm, ..., ip0:ip1] = vxc_nldf * weight
                den_a = rho_a[0] * weight
                den_b = rho_b[0] * weight
                nelec[0, i] += den_a.sum()
                nelec[1, i] += den_b.sum()
                excsum[i] += np.dot(den_a, exc)
                excsum[i] += np.dot(den_b, exc)
                wva_full[idm, :, ip0:ip1] = weight * vxc[0]
                wvb_full[idm, :, ip0:ip1] = weight * vxc[1]
            ip0 = ip1
        for idm in range(nset):
            wva_full[idm, :, :] += ni.nldfgen.get_potential(
                vxc_nldf_full[idm, 0], spin=0
            )
            wvb_full[idm, :, :] += ni.nldfgen.get_potential(
                vxc_nldf_full[idm, 1], spin=1
            )

    buffers = None
    pair_mask = mol.get_overlap_cond() < -np.log(ni.cutoff)
    if any(x in xc_code.upper() for x in ("CC06", "CS", "BR89", "MK00")):
        raise NotImplementedError("laplacian in meta-GGA method")
    ao_deriv = 1
    v1 = np.zeros_like(vmat)
    for i, ip0, ip1, ao, mask, weight, coords in block_loop(ao_deriv):
        wva = wva_full[i, :, ip0:ip1]
        wvb = wvb_full[i, :, ip0:ip1]
        buffers = ni.contract_wv(
            ao,
            wva,
            nbins,
            mask,
            pair_mask,
            ao_loc,
            vmats=(vmat[0, i], v1[0, i]),
            buffers=buffers,
        )[1]
        buffers = ni.contract_wv(
            ao,
            wvb,
            nbins,
            mask,
            pair_mask,
            ao_loc,
            vmats=(vmat[1, i], v1[1, i]),
            buffers=buffers,
        )[1]

    vmat = lib.hermi_sum(vmat.reshape(-1, nao, nao), axes=(0, 2, 1)).reshape(
        2, nset, nao, nao
    )
    vmat += v1

    if ni.has_sdmx:
        ni.sdmxgen._cached_ao_data = None

    if isinstance(dma, np.ndarray) and is_2d:
        vmat = vmat[:, 0]
        nelec = nelec.reshape(2)
        excsum = excsum[0]

    dtype = np.result_type(dma, dmb)
    if vmat.dtype != dtype:
        vmat = np.asarray(vmat, dtype=dtype)
    ni.timer.stop("nr_uks")
    return nelec, excsum, vmat


class CiderNumIntMixin:

    sl_plan: SemilocalPlan
    fl_plan: FracLaplPlan
    sdmxgen: EXXSphGenerator
    nldfgen: PyscfNLDFGenerator
    settings: FeatureSettings
    xmix: float
    rhocut: float

    def nlc_coeff(self, xc_code):
        if self._nlc_coeff is not None:
            return self._nlc_coeff
        else:
            return super().nlc_coeff(xc_code)

    @property
    def settings(self):
        return self.mlxc.settings

    @property
    def has_sdmx(self):
        return self.settings.has_sdmx

    @property
    def has_nldf(self):
        return self.settings.has_nldf

    def build(self, mol=None):
        self.mol = mol
        self.timer = Timer()
        self.sl_plan = None
        self.fl_plan = None
        self.sdmxgen = None
        self.nldfgen = None

    def reset(self, mol=None):
        self.mol = mol
        self.sl_plan = None
        self.fl_plan = None
        self.sdmxgen = None
        self.nldfgen = None

    def initialize_feature_generators(self, mol, grids, nspin):
        self.sl_plan = SemilocalPlan(self.settings.sl_settings, nspin)
        self.fl_plan = FracLaplPlan(self.settings.nlof_settings, nspin)
        cond = self.sdmxgen is None
        cond = cond or self.mol != mol
        cond = cond or self.sdmxgen.plan.nspin != nspin
        cond = cond and self.settings.has_sdmx
        if cond:
            self.sdmxgen = self.sdmx_init.initialize_sdmx_generator(mol, nspin)
        self.mol = mol

    def eval_xc_cider(
        self,
        xc_code,
        rho,
        nldf_feat,
        sdmx_feat,
        deriv=1,
        omega=None,
        xctype=None,
        verbose=None,
    ):
        rho = np.asarray(rho, order="C", dtype=np.double)
        if deriv > 1:
            raise NotImplementedError
        if rho.ndim == 2:
            closed = True
            rho = rho[None, :]
        else:
            closed = False
        nfeat = self.settings.nfeat
        ngrids = rho.shape[-1]
        nspin = rho.shape[0]
        vxc = np.zeros_like(rho)
        xctype = self._xc_type(xc_code)
        if xctype == "LDA":
            nvar = 1
        elif xctype == "GGA":
            nvar = 4
        elif xctype == "MGGA":
            nvar = 5
        else:
            exc = np.zeros(ngrids)
        if xctype in ["LDA", "GGA", "MGGA"]:
            if nspin == 1:
                rhotmp = np.ascontiguousarray(rho[0, :nvar])
                exc, vxc[0, :nvar] = self.eval_xc_eff(
                    xc_code,
                    rhotmp,
                    deriv=deriv,
                    omega=omega,
                    xctype=xctype,
                    verbose=verbose,
                )[:2]
            else:
                exc, vxc[:, :nvar] = self.eval_xc_eff(
                    xc_code,
                    rho[:, :nvar],
                    deriv=deriv,
                    omega=omega,
                    xctype=xctype,
                    verbose=verbose,
                )[:2]

        start = 0
        X0T = np.empty((nspin, nfeat, ngrids))
        has_sl = not self.settings.sl_settings.is_empty
        has_nldf = not self.settings.nldf_settings.is_empty
        has_nlof = not self.settings.nlof_settings.is_empty
        has_sdmx = not self.settings.sdmx_settings.is_empty
        if has_sl:
            nfeat_tmp = self.settings.sl_settings.nfeat
            X0T[:, start : start + nfeat_tmp] = self.sl_plan.get_feat(rho[:, :5])
            start += nfeat_tmp
        if has_nldf:
            if nldf_feat.ndim == 2:
                nldf_feat = nldf_feat[None, :]
            assert nldf_feat.shape[0] == nspin
            assert nldf_feat.ndim == 3
            nfeat_tmp = self.settings.nldf_settings.nfeat
            X0T[:, start : start + nfeat_tmp] = nldf_feat
            start += nfeat_tmp
        if has_nlof:
            nfeat_tmp = self.settings.nlof_settings.nfeat
            X0T[:, start : start + nfeat_tmp] = self.fl_plan.get_feat(rho)
            start += nfeat_tmp
        if has_sdmx:
            if sdmx_feat.ndim == 2:
                sdmx_feat = sdmx_feat[None, :]
            assert sdmx_feat.shape[0] == nspin
            assert sdmx_feat.ndim == 3
            nfeat_tmp = self.settings.sdmx_settings.nfeat
            X0T[:, start : start + nfeat_tmp] = sdmx_feat
            start += nfeat_tmp
        if start != nfeat:
            raise RuntimeError("nfeat mismatch, this should not happen!")

        X0TN = self.settings.normalizers.get_normalized_feature_vector(X0T)
        if isinstance(self.mlxc, MappedXC):
            exc_ml, dexcdX0TN_ml = self.mlxc(X0TN, rhocut=self.rhocut)
        elif isinstance(self.mlxc, MappedXC2):
            rho_tuple = get_rho_tuple_with_grad_cross(rho, is_mgga=True)
            exc_ml, dexcdX0TN_ml, vrho_tuple = self.mlxc(
                X0TN, rho_tuple, rhocut=self.rhocut
            )
            vxc[:] += vxc_tuple_to_array(rho, vrho_tuple)
        else:
            raise TypeError("mlxc must be MappedXC or MappedXC2")
        xmix = self.xmix  # / rho.shape[0]
        exc_ml *= xmix
        dexcdX0TN_ml *= xmix
        vxc_ml = self.settings.normalizers.get_derivative_wrt_unnormed_features(
            X0T, dexcdX0TN_ml
        )
        exc[:] += exc_ml / (rho[:, 0].sum(axis=0) + 1e-16)

        start = 0
        if has_sl:
            nfeat_tmp = self.settings.sl_settings.nfeat
            self.sl_plan.get_vxc(
                rho[:, :5], vxc_ml[:, start : start + nfeat_tmp], vxc=vxc[:, :5]
            )
            start += nfeat_tmp
        if has_nldf:
            nfeat_tmp = self.settings.nldf_settings.nfeat
            vxc_nldf = vxc_ml[:, start : start + nfeat_tmp]
            start += nfeat_tmp
        else:
            vxc_nldf = None
        if has_nlof:
            nfeat_tmp = self.settings.nlof_settings.nfeat
            self.fl_plan.get_vxc(vxc_ml[:, start : start + nfeat_tmp], vxc=vxc)
            start += nfeat_tmp
        if has_sdmx:
            nfeat_tmp = self.settings.sdmx_settings.nfeat
            vxc_sdmx = vxc_ml[:, start : start + nfeat_tmp]
            start += nfeat_tmp
        else:
            vxc_sdmx = None
        if start != nfeat:
            raise RuntimeError("nfeat mismatch, this should not happen!")

        if closed:
            vxc = vxc[0]

        return exc, (vxc, vxc_nldf, vxc_sdmx), None, None


class CiderNumInt(CiderNumIntMixin, numint.NumInt):
    def __init__(
        self, mlxc, nldf_init, sdmx_init, xmix=1.0, rhocut=None, nlc_coeff=None
    ):
        """

        Args:
            mlxc (MappedXC): Model for XC energy
            nldf_init (PySCFNLDFInitializer)
            sdmx_init (PySCFSDMXInitializer)
            xmix (float): Mixing fraction of ML functional
            rhocut (float): Low density cutoff for numerical stability
        """
        self.mlxc = mlxc
        self.xmix = xmix
        self.rhocut = DEFAULT_RHOCUT if rhocut is None else rhocut
        self.mol = None
        self.nldf_init = nldf_init
        self.sdmx_init = sdmx_init
        self.sdmxgen = None
        self.nldfgen = None
        self._nlc_coeff = nlc_coeff
        super(CiderNumInt, self).__init__()

    def method_not_implemented(self, *args, **kwargs):
        raise NotImplementedError

    nr_rks = nr_rks
    nr_uks = nr_uks

    nr_rks_fxc = method_not_implemented
    nr_uks_fxc = method_not_implemented
    nr_rks_fxc_st = method_not_implemented
    cache_xc_kernel = method_not_implemented
    cache_xc_kernel1 = method_not_implemented

    def contract_wv(
        self,
        ao,
        wv,
        nbins,
        mask,
        pair_mask,
        ao_loc,
        buffers=None,
        vmats=None,
    ):
        if vmats is None:
            vmats = (None, None)
        if not wv.flags.c_contiguous:
            wv = np.ascontiguousarray(wv)
        vmat, v1 = vmats
        aow = buffers
        wv[0] *= 0.5
        aow = _scale_ao_sparse(ao[:4], wv[:4], mask, ao_loc, out=aow)
        vmat = _dot_ao_ao_sparse(
            ao[0], aow, None, nbins, mask, pair_mask, ao_loc, hermi=0, out=vmat
        )
        if len(wv) > 4:
            wv[4] *= 0.5
            v1 = _tau_dot_sparse(
                ao, ao, wv[4], nbins, mask, pair_mask, ao_loc, out=v1, aow=aow
            )
        return (vmat, v1), aow

    def block_loop(
        self,
        mol,
        grids,
        nao=None,
        deriv=0,
        max_memory=2000,
        non0tab=None,
        blksize=None,
        buf=None,
        extra_ao=0,
    ):
        """Define this macro to loop over grids by blocks.
        Same as pyscf block_loop except extra_ao allows accounting for
        extra ao terms in blksize calculation.
        """
        if grids.coords is None:
            grids.build(with_non0tab=True)
        if nao is None:
            nao = mol.nao
        ngrids = grids.coords.shape[0]
        comp = (deriv + 1) * (deriv + 2) * (deriv + 3) // 6
        # NOTE to index grids.non0tab, the blksize needs to be an integer
        # multiplier of BLKSIZE
        if blksize is None:
            blksize = int(
                max_memory * 1e6 / (((comp + 1) * nao + extra_ao) * 8 * BLKSIZE)
            )
            blksize = max(4, min(blksize, ngrids // BLKSIZE + 1, 1200)) * BLKSIZE
        assert blksize % BLKSIZE == 0

        if non0tab is None and mol is grids.mol:
            non0tab = grids.non0tab
        if non0tab is None:
            non0tab = np.empty(
                ((ngrids + BLKSIZE - 1) // BLKSIZE, mol.nbas), dtype=np.uint8
            )
            non0tab[:] = NBINS + 1  # Corresponding to AO value ~= 1
        screen_index = non0tab

        # the xxx_sparse() functions require ngrids 8-byte aligned
        allow_sparse = ngrids % ALIGNMENT_UNIT == 0 and nao > numint.SWITCH_SIZE

        if buf is None:
            buf = numint._empty_aligned(comp * blksize * nao)
        for ip0, ip1 in lib.prange(0, ngrids, blksize):
            coords = grids.coords[ip0:ip1]
            weight = grids.weights[ip0:ip1]
            mask = screen_index[ip0 // BLKSIZE :]
            # TODO: pass grids.cutoff to eval_ao
            ao = self.eval_ao(
                mol, coords, deriv=deriv, non0tab=mask, cutoff=grids.cutoff, out=buf
            )
            if not allow_sparse and not numint._sparse_enough(mask):
                # Unset mask for dense AO tensor. It determines which eval_rho
                # to be called in make_rho
                mask = None
            yield ao, mask, weight, coords


class _FLNumIntMixin:

    feat_plan = None
    settings: FeatureSettings = None

    nr_rks = nr_rks
    nr_uks = nr_uks

    nr_nlc_vxc = numint.nr_nlc_vxc
    nr_sap = nr_sap_vxc = numint.nr_sap_vxc

    make_mask = staticmethod(nlof.make_mask)
    eval_ao = staticmethod(numint.eval_ao)
    eval_kao = staticmethod(nlof.eval_kao)
    eval_rho = staticmethod(nlof.eval_rho)
    eval_rho1 = lib.module_method(nlof.eval_rho1, absences=["cutoff"])
    eval_rho2 = staticmethod(nlof.eval_rho2)
    get_rho = numint.get_rho

    def block_loop(
        ni,
        mol,
        grids,
        nao=None,
        deriv=0,
        max_memory=2000,
        non0tab=None,
        blksize=None,
        buf=None,
        extra_ao=None,
    ):
        """Define this macro to loop over grids by blocks. Expands on
        CiderNumInt.block_loop by also returning the fractional laplacian-
        like operators on the orbitals, as set up by the FracLaplSettings
        """
        if grids.coords is None:
            grids.build(with_non0tab=True)
        if nao is None:
            nao = mol.nao
        ngrids = grids.coords.shape[0]
        comp = (deriv + 1) * (deriv + 2) * (deriv + 3) // 6
        flsettings = ni.settings.nlof_settings
        kcomp = flsettings.npow + 3 * flsettings.nd1
        # NOTE to index grids.non0tab, the blksize needs to be an integer
        # multiplier of BLKSIZE
        if blksize is None:
            blksize = int(
                max_memory * 1e6 / (((comp + kcomp + 1) * nao + extra_ao) * 8 * BLKSIZE)
            )
            blksize = max(4, min(blksize, ngrids // BLKSIZE + 1, 1200)) * BLKSIZE
        assert blksize % BLKSIZE == 0

        if non0tab is None and mol is grids.mol:
            non0tab = grids.non0tab
        if non0tab is None:
            non0tab = np.empty(
                ((ngrids + BLKSIZE - 1) // BLKSIZE, mol.nbas), dtype=np.uint8
            )
            non0tab[:] = NBINS + 1  # Corresponding to AO value ~= 1
        screen_index = non0tab

        # the xxx_sparse() functions require ngrids 8-byte aligned
        allow_sparse = ngrids % ALIGNMENT_UNIT == 0

        if buf is None:
            buf = nlof._empty_aligned((comp + kcomp) * blksize * nao)
        for ip0, ip1 in lib.prange(0, ngrids, blksize):
            coords = grids.coords[ip0:ip1]
            weight = grids.weights[ip0:ip1]
            mask = screen_index[ip0 // BLKSIZE :]
            # TODO: pass grids.cutoff to eval_ao
            ao_buf = np.ndarray((comp + kcomp, weight.size * nao), buffer=buf)
            ao = ni.eval_ao(
                mol, coords, deriv=deriv, non0tab=mask, cutoff=grids.cutoff, out=ao_buf
            )
            kao = ni.eval_kao(
                flsettings.slist,
                mol,
                coords,
                deriv=0,
                non0tab=None,
                cutoff=grids.cutoff,
                out=ao_buf[comp:],
                n1=flsettings.nd1,
            )
            if not allow_sparse and not nlof._sparse_enough(mask):
                # Unset mask for dense AO tensor. It determines which eval_rho
                # to be called in make_rho
                mask = None
            yield (ao, kao), mask, weight, coords

    def contract_wv(
        self,
        ao,
        wv,
        nbins,
        mask,
        pair_mask,
        ao_loc,
        buffers=None,
        vmats=None,
    ):
        if buffers is None:
            buffers = (None, None)
        if vmats is None:
            vmats = (None, None)
        aow1, aow2, vmat, v1 = nlof._odp_dot_sparse_(
            ao,
            wv,
            nbins,
            mask,
            pair_mask,
            ao_loc,
            self.settings.nlof_settings,
            aow1=buffers[0],
            aow2=buffers[1],
            vmat=vmats[0],
            v1=vmats[1],
        )
        return (vmat, v1), (aow1, aow2)

    _gen_rho_evaluator = nlof.FLNumInt._gen_rho_evaluator


class _NLDFMixin:

    nr_rks = nr_rks_nldf
    nr_uks = nr_uks_nldf

    def extra_block_loop(
        self,
        mol,
        grids,
        max_memory=2000,
        non0tab=None,
        blksize=None,
        make_mask=False,
        extra_ao=None,
    ):
        if grids.coords is None:
            grids.build(with_non0tab=True)
        ngrids = grids.coords.shape[0]
        # NOTE to index grids.non0tab, the blksize needs to be an integer
        # multiplier of BLKSIZE
        if blksize is None:
            blksize = int(max_memory * 1e6 / ((1 + extra_ao) * 8 * BLKSIZE))
            blksize = max(4, min(blksize, ngrids // BLKSIZE + 1, 1200)) * BLKSIZE
        assert blksize % BLKSIZE == 0

        if make_mask:
            if non0tab is None and mol is grids.mol:
                non0tab = grids.non0tab
            if non0tab is None:
                non0tab = np.empty(
                    ((ngrids + BLKSIZE - 1) // BLKSIZE, mol.nbas), dtype=np.uint8
                )
                non0tab[:] = NBINS + 1  # Corresponding to AO value ~= 1
            screen_index = non0tab
        else:
            screen_index = None

        for ip0, ip1 in lib.prange(0, ngrids, blksize):
            coords = grids.coords[ip0:ip1]
            weight = grids.weights[ip0:ip1]
            if make_mask:
                mask = screen_index[ip0 // BLKSIZE :]
            else:
                mask = None
            yield mask, weight, coords


class NLOFNumInt(_FLNumIntMixin, CiderNumInt):
    pass


class NLDFNumInt(_NLDFMixin, CiderNumInt):

    grids = None

    def initialize_feature_generators(self, mol, grids, nspin):
        cond = self.nldfgen is None
        cond = cond or self.grids != grids
        cond = cond or self.mol != mol
        cond = cond or self.nldfgen.plan.nspin != nspin
        if cond:
            self.nldfgen = self.nldf_init.initialize_nldf_generator(
                mol, grids.grids_indexer, nspin
            )
            self.nldfgen.interpolator.set_coords(grids.coords)
        super().initialize_feature_generators(mol, grids, nspin)
        self.grids = grids


class NLDFNLOFNumInt(_NLDFMixin, _FLNumIntMixin, CiderNumInt):

    grids = None

    def initialize_feature_generators(self, mol, grids, nspin):
        cond = self.nldfgen is None
        cond = cond or self.grids != grids
        cond = cond or self.mol != mol
        cond = cond or self.nldfgen.plan.nspin != nspin
        if cond:
            self.nldfgen = self.nldf_init.initialize_nldf_generator(
                mol, grids.grids_indexer, nspin
            )
        super().initialize_feature_generators(mol, grids, nspin)
        self.grids = grids
